""" Message purging plugin. """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from typing import ClassVar, Optional

from pyrogram.errors import MessageDeleteForbidden

from anjani import command, filters, plugin


class Purges(plugin.Plugin):
    name: ClassVar[str] = "Purges"
    helpable: ClassVar[bool] = True

    @command.filters(filters.can_delete)
    async def cmd_del(self, ctx: command.Context) -> Optional[str]:
        """Delete replied message"""
        reply_msg = ctx.msg.reply_to_message
        if not reply_msg:
            return await self.text(ctx.chat.id, "error-reply-to-message")

        try:
            await self.bot.client.delete_messages(ctx.chat.id, [reply_msg.id, ctx.msg.id])
        except MessageDeleteForbidden as e:
            await ctx.respond(
                await self.text(ctx.chat.id, "purge-failed", e.MESSAGE), delete_after=5
            )

    @command.filters(filters.can_delete, aliases=["prune"])
    async def cmd_purge(self, ctx: command.Context) -> Optional[str]:
        """purge message from message replied"""
        if not ctx.msg.reply_to_message:
            return await self.text(ctx.chat.id, "error-reply-to-message")

        time_start = datetime.now()
        start, end = ctx.msg.reply_to_message.id, ctx.msg.id
        messages = [*range(start, end)]

        try:
            await self.bot.client.delete_messages(chat_id=ctx.chat.id, message_ids=messages)
        except MessageDeleteForbidden:
            await ctx.respond(await self.text(ctx.chat.id, "purge-error", delete_after=5))
            return None
        else:
            await ctx.msg.delete()

        time_end = datetime.now()
        run_time = (time_end - time_start).seconds

        await ctx.respond(
            await self.text(ctx.chat.id, "purge-done", len(messages), run_time), delete_after=5
        )
        return None



"""Anjani base chat action"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

from pyrogram.enums.chat_action import ChatAction
from pyrogram.errors import FloodWait
from pyrogram.types import Chat

if TYPE_CHECKING:
    from anjani.command import Context
    from anjani.core import Anjani


class BotAction:

    # Instances variable
    __running: bool
    __current: ChatAction
    __chat: Chat

    bot: "Anjani"
    loop: asyncio.AbstractEventLoop

    # Instance variable to be filled later
    __task: asyncio.Task[None]

    def __init__(self, ctx: "Context", action: ChatAction = ChatAction.TYPING) -> None:
        self.__running = True
        self.__current = action
        self.__chat = ctx.chat

        self.bot = ctx.bot
        self.loop = ctx.bot.loop

    async def __cancel(self) -> None:
        try:
            await self.bot.client.send_chat_action(self.__chat.id, ChatAction.CANCEL)
        except FloodWait as e:
            await asyncio.sleep(e.value)  # type: ignore

    async def __start(self) -> None:
        while self.__running:
            try:
                await self.bot.client.send_chat_action(self.__chat.id, self.__current)
            except FloodWait as e:
                await asyncio.sleep(e.value)  # type: ignore
            else:
                await asyncio.sleep(1)

    async def __stop(self) -> None:
        self.__running = False
        await self.__cancel()

        if not self.__task.done():
            self.__task.cancel()
        else:
            self.__task.result()

    async def switch(self, action: ChatAction) -> None:
        """Switch current BotAction"""
        # avoid race condition with current action
        async with asyncio.Lock():
            await self.__cancel()
            self.__current = action

    def __enter__(self) -> "BotAction":
        self.__task = self.loop.create_task(self.__start())
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc: Optional[Exception],
        tb: Optional[TracebackType],
    ) -> None:
        self.loop.create_task(self.__stop())

    async def __aenter__(self) -> "BotAction":
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc: Optional[Exception],
        tb: Optional[TracebackType],
    ) -> None:
        await self.__stop()



from os import cpu_count, getenv
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv

from anjani import DEFAULT_CONFIG_PATH


class Config:
    API_ID: str
    API_HASH: str
    BOT_TOKEN: str
    OWNER_ID: int
    WORKERS: int
    DOWNLOAD_PATH: Optional[str]

    DB_URI: str

    SW_API: Optional[str]
    LOG_CHANNEL: Optional[str]
    ALERT_LOG: Optional[str]

    LOGIN_URL: Optional[str]
    PLUGIN_FLAG: list[str]
    FEATURE_FLAG: list[str]

    HEALTH_CHECK_INTERVAL: Optional[int]
    HEALTH_CHECK_WEBHOOK_URL: Optional[str]

    SPAM_PREDICTION_URL: Optional[str]
    SPAM_PREDICTION_API: Optional[str]

    IS_CI: bool

    def __init__(self) -> None:
        config_path = Path(DEFAULT_CONFIG_PATH)
        if config_path.is_file():
            load_dotenv(config_path)

        self.API_ID = getenv("API_ID", "")
        self.API_HASH = getenv("API_HASH", "")
        self.BOT_TOKEN = getenv("BOT_TOKEN", "")
        self.OWNER_ID = int(getenv("OWNER_ID", 0))
        self.WORKERS = int(getenv("WORKERS", min(32, (cpu_count() or 0) + 4)))
        self.DOWNLOAD_PATH = getenv("DOWNLOAD_PATH", "./downloads")

        self.DB_URI = getenv("DB_URI", "")

        self.LOG_CHANNEL = getenv("LOG_CHANNEL")
        self.ALERT_LOG = getenv("ALERT_LOG")
        self.SW_API = getenv("SW_API")

        self.LOGIN_URL = getenv("LOGIN_URL")
        self.PLUGIN_FLAG = list(
            filter(None, [i.strip() for i in getenv("PLUGIN_FLAG", "").split(";")])
        )
        self.FEATURE_FLAG = list(
            filter(None, [i.strip() for i in getenv("FEATURE_FLAG", "").split(";")])
        )

        self.HEALTH_CHECK_INTERVAL = int(getenv("HEALTH_CHECK_INTERVAL", 60))
        self.HEALTH_CHECK_WEBHOOK_URL = getenv("HEALTH_CHECK_WEBHOOK_URL")

        self.SPAM_PREDICTION_URL = getenv("SPAM_PREDICTION_URL")
        self.SPAM_PREDICTION_API = getenv("SPAM_PREDICTION_API")

        self.IS_CI = getenv("IS_CI", "false").lower() == "true"

        #  check if all the required variables are set
        if any(
            {
                not self.API_ID,
                not self.API_HASH,
                not self.BOT_TOKEN,
                not self.DB_URI,
            }
        ):
            raise RuntimeError("Required ENV variables are missing!")

        # create download path if not exists
        Path(self.DOWNLOAD_PATH).mkdir(parents=True, exist_ok=True)

    def is_plugin_disabled(self, name: str) -> bool:
        return f'disable_{name.lower().replace(" ", "_")}_plugin' in self.PLUGIN_FLAG

    def is_flag_active(self, name: str) -> bool:
        return name in self.FEATURE_FLAG



""" Canonical plugin for @dAnjani_bot """

# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
from base64 import b64encode
from typing import Any, ClassVar, MutableMapping

from aiohttp import web
from aiopath import AsyncPath
from prometheus_client import REGISTRY, generate_latest
from pymongo.errors import PyMongoError
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_members_filter import ChatMembersFilter
from pyrogram.enums.message_media_type import MessageMediaType
from pyrogram.types import (
    ChatMemberUpdated,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    LoginUrl,
    Message,
)

from anjani import command, filters, listener, plugin
from anjani.core.metrics import MessageStat

# metrics endpoint filter
logging.getLogger("aiohttp.access").setLevel(logging.WARNING)

async def metrics_handler(_: web.Request):
    metrics = generate_latest(REGISTRY)
    return web.Response(body=metrics, content_type="text/plain")


class Canonical(plugin.Plugin):
    """Helper Plugin
    This plugin is only available for @dAnjani_bot
    to comunicate with https://userbotindo.com
    """

    name: ClassVar[str] = "Canonical"

    # Private
    _web_runner: web.AppRunner
    _web_site: web.TCPSite

    __web_task: asyncio.Task[None]
    __task: asyncio.Task[None]
    _mt: MutableMapping[MessageMediaType, str] = {
        MessageMediaType.STICKER: "sticker",
        MessageMediaType.PHOTO: "photo",
        MessageMediaType.DOCUMENT: "file",
        MessageMediaType.VIDEO: "video",
    }

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("TEST")
        self.chats_db = self.bot.db.get_collection("CHATS")
        await self._setup_web_app()

    async def _setup_web_app(self):
        app = web.Application()
        app.add_routes([web.get("/metrics", metrics_handler)])
        self._web_runner = web.AppRunner(app)
        await self._web_runner.setup()
        self._web_site = web.TCPSite(self._web_runner, "0.0.0.0", 9090)
        await self._web_site.start()

    async def stop_aiohttp_server(self):
        if self._web_site:
            await self._web_site.stop()
        if self._web_runner:
            await self._web_runner.cleanup()

    async def on_start(self, _: int) -> None:
        self.log.debug("Starting watch streams")
        self.__task = self.bot.loop.create_task(self.watch_streams())
        self.__web_task = self.bot.loop.create_task(self._setup_web_app())

        async def _web_shutdown(task: asyncio.Task[None]) -> None:
            if task.cancelled():
                await self.stop_aiohttp_server()

        def shutdown_wrapper(task):
            asyncio.create_task(_web_shutdown(task))

        self.__web_task.add_done_callback(shutdown_wrapper)

    async def on_stop(self) -> None:
        self.log.debug("Stopping watch streams")
        self.__task.cancel()
        self.log.debug("Shutting down web app")
        self.__web_task.cancel()

    def get_type(self, message: Message) -> str:
        if message.command:
            return "command"
        return self._mt.get(message.media, "text") if message.media else "text"

    @command.filters(filters.admin_only & filters.group)
    async def cmd_r(self, ctx: command.Context) -> None:
        """Refresh chat data"""
        admins = []
        async for member in self.bot.client.get_chat_members(
            ctx.chat.id, filter=ChatMembersFilter.ADMINISTRATORS
        ):  # type: ignore
            if not member.user.is_bot and member.privileges.can_manage_chat:
                admins.append(member.user.id)

        photo = None
        if ctx.chat.photo:
            file = await self.bot.client.download_media(ctx.chat.photo.small_file_id)
            if file and isinstance(file, str):
                file = AsyncPath(file)
                result = await file.read_bytes()
                photo = b64encode(result).decode("utf-8")
                await file.unlink()
            if file and isinstance(file, bytes):
                photo = b64encode(file).decode("utf-8")

        await self.chats_db.update_one(
            {"chat_id": ctx.chat.id}, {"$set": {"admins": admins, "photo": photo}}, upsert=True
        )
        await ctx.respond("Done", delete_after=5)

    @listener.priority(65)
    async def on_message(self, message: Message) -> None:
        """Message metric analytics"""
        if message.outgoing:
            return

        MessageStat.labels(self.get_type(message)).inc()

    async def on_chat_action(self, message: Message) -> None:
        """Delete admins data from chats"""
        if message.new_chat_members:
            return

        chat = message.chat
        user = message.left_chat_member
        if user.id == self.bot.uid:
            await asyncio.gather(
                self.chats_db.update_one({"chat_id": chat.id}, {"$set": {"admins": []}}),
            )

    async def on_chat_member_update(self, update: ChatMemberUpdated) -> None:
        old_data = update.old_chat_member
        new_data = update.new_chat_member

        if not old_data or not new_data:  # Rare case
            return

        if old_data.status == new_data.status:
            return

        if new_data.status not in {
            ChatMemberStatus.ADMINISTRATOR,
            ChatMemberStatus.OWNER,
        }:
            await self.chats_db.update_one(
                {"chat_id": update.chat.id}, {"$pull": {"admins": new_data.user.id}}
            )
        elif (
            new_data.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}
            and new_data.privileges
            and new_data.privileges.can_manage_chat
        ):  # type: ignore
            await self.chats_db.update_one(
                {"chat_id": update.chat.id},
                {"$addToSet": {"admins": new_data.user.id}},
                upsert=True,
            )

    async def watch_streams(self) -> None:
        try:
            async with self.db.watch([{"$match": {"operationType": "insert"}}]) as stream:
                async for change in stream:
                    await self.dispatch_change(change["fullDocument"])
        except PyMongoError as e:
            self.log.error("Error", exc_info=e)

    async def dispatch_change(self, doc: MutableMapping[str, Any]) -> None:
        chat_id = int(doc["_id"])
        message = doc["message"]
        pin = doc.get("pin", False)
        disable_preview = doc.get("disable_preview", False)

        try:
            msg = await self.bot.client.send_message(
                chat_id=chat_id,
                text=message,
                disable_web_page_preview=disable_preview,
            )
            if pin:
                await self.bot.client.pin_chat_message(msg.chat.id, msg.id)
        except Exception as e:  # skipcq: PYL-W0703
            self.log.error(f"Error sending message to {chat_id}", exc_info=e)
        finally:
            await self.db.delete_one({"_id": chat_id})

    @command.filters(filters.private)
    async def cmd_login(self, ctx: command.Context):
        """Login to https://userbotindo.com"""
        if not self.bot.config.LOGIN_URL:
            return

        await ctx.respond(
            "Click this button to login to Anjani Dashboard",
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            "Login",
                            login_url=LoginUrl(
                                url=self.bot.config.LOGIN_URL,
                                forward_text="Login to https://userbotindo.com",
                                request_write_access="True",
                            ),
                        )
                    ]
                ]
            ),
        )



"""Anjani database stream"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, List, Literal, Mapping, Optional, Union

from bson.timestamp import Timestamp
from pymongo.change_stream import ChangeStream
from pymongo.collation import Collation

from anjani import util

from .base import AsyncBase
from .client_session import AsyncClientSession

if TYPE_CHECKING:
    from .client import AsyncClient
    from .collection import AsyncCollection
    from .db import AsyncDatabase


class AsyncChangeStream(AsyncBase):
    """AsyncIO :obj:`~ChangeStream`

    *DEPRECATED* methods are removed in this class.
    """

    _target: Union["AsyncClient", "AsyncDatabase", "AsyncCollection"]

    dispatch: ChangeStream

    def __init__(
        self,
        target: Union["AsyncClient", "AsyncDatabase", "AsyncCollection"],
        pipeline: Optional[List[Mapping[str, Any]]],
        full_document: Optional[Literal["updateLookup"]],
        resume_after: Optional[Mapping[str, str]],
        max_await_time_ms: Optional[int],
        batch_size: Optional[int],
        collation: Optional[Collation],
        start_at_operation_time: Optional[Timestamp],
        session: Optional[AsyncClientSession],
        start_after: Optional[Mapping[str, str]],
        comment: Optional[str],
        full_document_before_change: Optional[Literal["required", "whenAvailable"]] = None,
    ) -> None:
        self._target = target
        self._options: Mapping[str, Any] = {
            "pipeline": pipeline,
            "full_document": full_document,
            "resume_after": resume_after,
            "max_await_time_ms": max_await_time_ms,
            "batch_size": batch_size,
            "collation": collation,
            "start_at_operation_time": start_at_operation_time,
            "session": session.dispatch if session else session,
            "start_after": start_after,
            "comment": comment,
            "full_document_before_change": full_document_before_change,
        }

        super().__init__(None)  # type: ignore

    def __aiter__(self) -> "AsyncChangeStream":
        return self

    def __iter__(self) -> None:
        raise RuntimeError("Use 'async for' instead of 'for'")

    async def __anext__(self) -> Mapping[str, Any]:
        return await self.next()

    async def __aenter__(self) -> "AsyncChangeStream":
        if not self.dispatch:
            await self._init()

        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.dispatch:
            await self.close()

    def __enter__(self) -> None:
        raise RuntimeError("Use 'async with' not just 'with'")

    async def _init(self) -> ChangeStream:
        if not self.dispatch:
            self.dispatch = await util.run_sync(self._target.dispatch.watch, **self._options)

        return self.dispatch

    async def close(self):
        if self.dispatch:
            await util.run_sync(self.dispatch.close)

    async def next(self) -> Mapping[str, Any]:
        while self.alive:
            document = await self.try_next()
            if document:
                return document

        raise StopAsyncIteration

    async def try_next(self) -> Optional[Mapping[str, Any]]:
        self.dispatch = await self._init()
        return await util.run_sync(self.dispatch.try_next)

    @property
    def alive(self) -> bool:
        if not self.dispatch:
            return True

        return self.dispatch.alive

    @property
    def resume_token(self) -> Any:
        if self.dispatch:
            return self.dispatch.resume_token

        return None



"""Anjani event listener"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Callable, Optional, Union

from pyrogram.filters import Filter

ListenerFunc = Callable[..., Any]
Decorator = Callable[[ListenerFunc], ListenerFunc]


def priority(_prio: int) -> Decorator:
    """Sets priority on the given listener function."""

    def prio_decorator(func: ListenerFunc) -> ListenerFunc:
        setattr(func, "_listener_priority", _prio)
        return func

    return prio_decorator


def filters(_filters: Filter) -> Decorator:
    """Sets filters on the given listener function."""

    def filters_decorator(func: ListenerFunc) -> ListenerFunc:
        setattr(func, "_listener_filters", _filters)
        return func

    return filters_decorator


class Listener:
    event: str
    func: Union[ListenerFunc, ListenerFunc]
    plugin: Any
    priority: int
    filters: Optional[Filter]

    def __init__(
        self,
        event: str,
        func: ListenerFunc,
        plugin: Any,
        prio: int,
        listener_filter: Optional[Filter] = None,
    ) -> None:
        self.event = event
        self.func = func
        self.plugin = plugin
        self.priority = prio
        self.filters = listener_filter

    def __lt__(self, other: "Listener") -> bool:
        return self.priority < other.priority

    def __repr__(self) -> str:
        return f"<listener plugin '{self.event}' from '{self.plugin.name}'>"



"""Anjani Errors Constructor"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, Optional, Type

if TYPE_CHECKING:
    from .command import Command
    from .plugin import Plugin
    from .util.converter import Converter

__all__ = [
    "AnjaniException",
    "BackupError",
    "BadArgument",
    "BadBoolArgument",
    "BadResult",
    "CommandHandlerError",
    "CommandInvokeError",
    "ConversionError",
    "EventDispatchError",
    "ExistingCommandError",
    "ExistingPluginError",
    "PluginLoadError",
]


class AnjaniException(Exception):
    """Base exception class for Anjani"""


class BackupError(AnjaniException):
    """Unexpected backup data type"""


class EventDispatchError(AnjaniException):
    """Exception raised when a dispatcher raised an exception"""


class CommandHandlerError(AnjaniException):
    """Exception raised when the command handler raised an exception."""


class CommandInvokeError(AnjaniException):
    """Exception raised when the command being invoked raised an exception."""


class PluginLoadError(AnjaniException):
    """Base exception class for every Plugin errors"""


class BadArgument(AnjaniException):
    """Exception that raised when parsing or converting argument encountered a failure."""


class ConversionError(BadArgument):
    """Base exception class for any argument conversion errors

    Attributes:
        converter (`Converter`): The converter instance that failed.
        origin (`Exception`): The original exception that was raised.
    """

    def __init__(  # skipcq: PYL-W1113
        self, converter: Optional["Converter"] = None, err: Optional[Exception] = None, *args: Any
    ) -> None:
        if converter and err:
            self.converter = type(converter).__name__
            self.origin = err
            super().__init__(f"Failed to convert on '{self.converter}', {err}")
        else:
            super().__init__(*args)


class BadBoolArgument(ConversionError):
    """Exception that raised when a bool argument can't be converted"""

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)


class BadResult(ConversionError):
    """Exception that raised when an argument results is not a valid types"""

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)


class ExistingCommandError(PluginLoadError):
    """Exception that raised when a command registered more then one.

    Attributes:
        old_cmd (:obj:`Command`): The old command that already registered.
        new_cmd (:obj:`Command`): The new command that already registered.
        alias (:obj:`bool`): Wether the command is an alias or not.
    """

    def __init__(self, old_cmd: "Command", new_cmd: "Command", alias: bool = False) -> None:
        al_str = "alias of " if alias else ""
        old_name = type(old_cmd.plugin).__name__
        new_name = type(new_cmd.plugin).__name__
        self.old_cmd = old_cmd
        self.new_cmd = new_cmd
        self.alias = alias
        super().__init__(
            f"Attempt to replace existing command '{old_cmd.name}' (from {old_name}) with {al_str}'{new_cmd.name}' (from {new_name})"
        )


class ExistingPluginError(PluginLoadError):
    """Exception that raised when two same Plugin name registered.

    Attributes:
        old_plugin (:obj:`Plugin`): The old plugin that already registered.
        new_plugin (:obj:`Plugin`): The new plugin that already registered.
        alias (:obj:`bool`): Wether the command is an alias or not.
    """

    def __init__(self, old_plugin: Type["Plugin"], new_plugin: Type["Plugin"]) -> None:
        self.old_plugin = old_plugin
        self.new_plugin = new_plugin
        super().__init__(f"Plugin '{old_plugin.name}' ({old_plugin.__name__}) already exists")



from aiocache.backends.memory import SimpleMemoryCache


class CacheLimiter(SimpleMemoryCache):
    # Initialized during instantiation
    max_value: int

    def __init__(self, ttl: int = 60, max_value: int = 10) -> None:
        super().__init__(ttl=ttl)

        self.max_value = max_value

    async def increment(self, user_id: int) -> None:  # skipcq: PYL-W0221
        """
        Increment the rate limit of the user
        """
        val = await self.get(user_id)
        await self.set(user_id, 1 if not val else val + 1, ttl=self.ttl)

    async def exceeded(self, user_id: int) -> bool:
        """
        Check if the user has exceeded the rate limit
        """
        val = await self.get(user_id)
        if not val:
            return False

        return val >= self.max_value



"""Anjani base command"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Coroutine,
    Iterable,
    Optional,
    Sequence,
    Union,
)

import pyrogram
from pyrogram.enums.chat_action import ChatAction
from pyrogram.filters import Filter
from pyrogram.types import Message
from typing_extensions import final

from anjani.action import BotAction
from anjani.util.tg import get_text
from anjani.util.types import CustomFilter

if TYPE_CHECKING:
    from anjani.core import Anjani

CommandFunc = Union[
    Callable[..., Coroutine[Any, Any, None]], Callable[..., Coroutine[Any, Any, Optional[str]]]
]
Decorator = Callable[[CommandFunc], CommandFunc]


def filters(_filters: Optional[Filter] = None, *, aliases: Iterable[str] = []) -> Decorator:
    """Sets filters on a command function."""

    def filter_decorator(func: CommandFunc) -> CommandFunc:
        setattr(func, "_cmd_filters", _filters)
        setattr(func, "_cmd_aliases", aliases)
        return func

    return filter_decorator


class Command:
    name: str
    plugin: Any
    func: Union[CommandFunc, CommandFunc]
    filters: Optional[Union[Filter, CustomFilter]]
    aliases: Iterable[str]

    def __init__(
        self,
        name: str,
        plugin: Any,
        func: CommandFunc,
        cmd_filter: Optional[Union[Filter, CustomFilter]],
        aliases: Iterable[str],
    ) -> None:
        self.name = name
        self.plugin = plugin
        self.func = func
        self.filters = cmd_filter
        self.aliases = aliases

    def __repr__(self) -> str:
        return f"<command plugin '{self.name}' from '{self.plugin.name}'>"


# Command invocation context
class Context:
    author: pyrogram.types.User
    bot: "Anjani"
    chat: pyrogram.types.Chat
    msg: Message
    message: Message
    cmd_len: int

    response: Message
    input: str
    input_raw: str
    args: Sequence[str]

    segments: Sequence[str]
    invoker: str

    def __init__(
        self,
        bot: "Anjani",
        msg: Message,
        cmd_len: int,
    ) -> None:
        self.bot = bot
        self.cmd_len = cmd_len
        self.msg = self.message = msg
        self.author = msg.from_user
        self.chat = msg.chat

        # Response message to be filled later
        self.response = None  # type: ignore
        # Single argument string
        username = self.bot.user.username
        if username and username in self.msg.text:
            slices = self.cmd_len + 1 + len(username)

            self.input = self.msg.text[slices:]
            self.input_raw = self.msg.text.markdown[slices:]
        else:
            self.input = self.msg.text[self.cmd_len :]
            self.input_raw = self.msg.text.markdown[self.cmd_len :]

        self.segments = self.msg.command
        self.invoker = self.segments[0]

    # Lazily resolve expensive fields
    def __getattr__(self, name: str) -> Any:
        if name == "args":
            return self._get_args()

        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    # Argument segments
    def _get_args(self) -> Sequence[str]:
        self.args = self.segments[1:]
        return self.args

    async def delete(
        self, delay: Optional[float] = None, message: Optional[Message] = None
    ) -> None:
        """Bound method of *delete* of :obj:`~pyrogram.types.Message`.
        If the deletion fails then it is silently ignored.

        delay (`float`, *optional*):
            If provided, the number of seconds to wait in the background
            before deleting the message.

        message (`~pyrogram.types.Message`, *optional*):
            If provided, the message passed will be deleted else will delete
            the client latest response.
        """
        content = message or self.response
        if not content:
            return

        if delay:

            async def delete(delay: float) -> None:
                await asyncio.sleep(delay)
                await content.delete(True)

            self.bot.loop.create_task(delete(delay))
        else:
            await content.delete(True)

    async def respond(
        self,
        text: str = "",
        *,
        animation: Optional[Union[str, BinaryIO, IO[bytes]]] = None,
        audio: Optional[Union[str, BinaryIO, IO[bytes]]] = None,
        document: Optional[Union[str, BinaryIO, IO[bytes]]] = None,
        photo: Optional[Union[str, BinaryIO, IO[bytes]]] = None,
        video: Optional[Union[str, BinaryIO, IO[bytes]]] = None,
        delete_after: Optional[Union[int, float]] = None,
        mode: str = "edit",
        redact: bool = True,
        reference: Optional[Message] = None,
        **kwargs: Any,
    ) -> Optional[Message]:
        """Respond to the destination with the content given.

        Parameters:
            text (`str`, *Optional*):
                Text of the message to be sent.
                *Optional* only if either :obj:`animation`, :obj:`audio`, :obj:`document`,
                :obj:`photo`, :obj:`video` or :obj:`delete_after` is provided.

            annimation (`str` | `BinaryIO`, *Optional*):
                Annimation to send. Pass a file_id as string to send an animation(GIF) file that
                exists on the Telegram servers, pass an HTTP URL as a string for Telegram
                to get an animation file from the Internet, pass a file path as string to upload
                a new animation file that exists on your local machine, or pass a binary
                file-like object with its attribute “.name” set for in-memory uploads.

            audio (`str` | `BinaryIO`, *Optional*):
                Audio file to send. Pass a file_id as string to send an audio file that
                exists on the Telegram servers, pass an HTTP URL as a string for Telegram
                to get an audio file from the Internet, pass a file path as string to upload
                a new audio file that exists on your local machine, or pass a binary
                file-like object with its attribute “.name” set for in-memory uploads.

            document (`str` | `BinaryIO`, *Optional*):
                File to send. Pass a file_id as string to send a file that exists on the
                Telegram servers, pass an HTTP URL as a string for Telegram to get a file
                from the Internet, pass a file path as string to upload a new file that
                exists on your local machine, or pass a binary file-like object with its
                attribute “.name” set for in-memory uploads.

            photo (`str` | `BinaryIO`, *Optional*):
                Photo to send. Pass a file_id as string to send a photo that exists on the
                Telegram servers, pass an HTTP URL as a string for Telegram to get a photo
                from the Internet, pass a file path as string to upload a new photo that
                exists on your local machine, or pass a binary file-like object with its
                attribute “.name” set for in-memory uploads.

            video (`str` | `BinaryIO`, *Optional*):
                Video to send. Pass a file_id as string to send a video that exists on the
                Telegram servers, pass an HTTP URL as a string for Telegram to get a video
                from the Internet, or pass a file path as string to upload a new video that
                exists on your local machine.

            delete_after (`float`, *Optional*):
                If provided, the number of seconds to wait in the background before deleting
                the message we just sent. If the deletion fails, then it is silently ignored.

            mode (`str`, *Optional*):
                The mode that the client will respond. "edit" and "reply". defaults to "edit".

            redact (`bool`, *Optional*):
                Tells wether the text will be redacted from sensitive environment key.
                Defaults to `True`.

            reference (`pyrogram.types.Message`, *Optional*):
                Tells client which message to respond (message to edit or to reply based on mode).
        """
        self.response = await self.bot.respond(
            reference or self.msg,
            text,
            mode=mode,
            redact=redact,
            response=self.response,
            animation=animation,
            audio=audio,
            document=document,
            photo=photo,
            video=video,
            **kwargs,
        )
        if delete_after:
            await self.delete(delete_after)
            self.response = None  # type: ignore
        return self.response

    async def trigger_action(self, action: ChatAction = ChatAction.TYPING) -> bool:
        """Triggers a ChatAction on the invoked chat.
        A Shortcut for *bot.client.send_chat_action()*

        Parameters:
            action (`~pyrogram.types.ChatAction`, *Optional*):
                Type of action to broadcast.
        """
        if not self.chat:
            return False

        return await self.bot.client.send_chat_action(self.chat.id, action)

    def action(self, action: ChatAction = ChatAction.TYPING) -> BotAction:
        """Returns a context manager that allows you to send a chat action
        for an indefinite time.

        Parameters:
            action (`~pyrogram.types.ChatAction`, *Optional*):
                Type of action to broadcast.
        """
        return BotAction(self, action)

    @final
    def get_text(
        self, text_name: str, *args: Any, noformat: bool = False, **kwargs: Any
    ) -> Coroutine[Any, Any, str]:
        """Bound method *get_text* of :obj:`~Anjani.plugin.Plugin`.

        Parse the string with user language setting.

        If :obj:`~pyrogram.types.Chat` is None text will always return as 'en'.

        Parameters:
            text_name (`str`):
                String name to parse. The string is parsed from YAML documents.
            *args (`any`, *Optional*):
                One or more values that should be formatted and inserted in the string.
                The value should be in order based on the language string placeholder.
            noformat (`bool`, *Optional*):
                If True, the text returned will not be formated.
                Default to False.
            **kwargs (`any`, *Optional*):
                One or more keyword values that should be formatted and inserted in the string.
                based on the keyword on the language strings.
        """
        chat_id = self.chat.id if self.chat else None
        return get_text(self.bot, chat_id, text_name, *args, noformat=noformat, **kwargs)



"""Anjani event dispatcher"""

# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import bisect
from datetime import datetime
from hashlib import sha256
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, Optional, Tuple

from pyrogram import raw
from pyrogram.filters import Filter
from pyrogram.raw import functions
from pyrogram.types import CallbackQuery, InlineQuery, Message

from anjani import plugin, util
from anjani.error import EventDispatchError
from anjani.listener import Listener, ListenerFunc
from anjani.util.misc import StopPropagation

from .anjani_mixin_base import MixinBase
from .metrics import EventCount, EventLatencySecond, UnhandledError

if TYPE_CHECKING:
    from .anjani_bot import Anjani

EventType = (
    CallbackQuery,
    InlineQuery,
    Message,
)


def _get_event_data(event: Any) -> MutableMapping[str, Any]:
    if isinstance(event, Message):
        return {
            "chat_title": event.chat.title,
            "chat_id": event.chat.id,
            "user_name": event.from_user.first_name if event.from_user else event.sender_chat.title,
            "user_id": event.from_user.id if event.from_user else event.sender_chat.id,
            "input": event.text,
        }
    if isinstance(event, CallbackQuery):
        return {
            "chat_title": event.message.chat.title,
            "chat_id": event.message.chat.id,
            "user_name": event.from_user.first_name,
            "user_id": event.from_user.id,
            "input": event.data,
        }
    if isinstance(event, InlineQuery):
        return {
            "user_name": event.from_user.first_name,
            "user_id": event.from_user.id,
            "input": event.query,
        }
    return {}


def _unpack_args(args: Tuple[Any, ...]) -> str:
    """Unpack arguments into a string for logging purposes."""
    return ", ".join([str(arg) for arg in args])


class EventDispatcher(MixinBase):
    # Initialized during instantiation
    listeners: MutableMapping[str, MutableSequence[Listener]]

    def __init__(self: "Anjani", **kwargs: Any) -> None:
        # Initialize listener map
        self.listeners = {}

        # Propagate initialization to other mixins
        super().__init__(**kwargs)

    def register_listener(
        self: "Anjani",
        plug: plugin.Plugin,
        event: str,
        func: ListenerFunc,
        *,
        priority: int = 100,
        filters: Optional[Filter] = None,
    ) -> None:
        if event in {"load", "start", "started", "stop", "stopped"} and filters is not None:
            self.log.warning("Built-in Listener can't be use with filters. Removing...")
            filters = None

        if getattr(func, "_cmd_filters", None):
            self.log.warning(
                "@command.filters decorator only for CommandFunc. Filters will be ignored..."
            )

        if filters:
            self.log.debug("Registering filter '%s' into '%s'", type(filters).__name__, event)

        listener = Listener(event, func, plug, priority, filters)

        if event in self.listeners:
            bisect.insort(self.listeners[event], listener)
        else:
            self.listeners[event] = [listener]

        self.update_plugin_events()

    def unregister_listener(self: "Anjani", listener: Listener) -> None:
        self.listeners[listener.event].remove(listener)
        # Remove list if empty
        if not self.listeners[listener.event]:
            del self.listeners[listener.event]

        self.update_plugin_events()

    def register_listeners(self: "Anjani", plug: plugin.Plugin) -> None:
        for event, func in util.misc.find_prefixed_funcs(plug, "on_"):
            done = True
            try:
                self.register_listener(
                    plug,
                    event,
                    func,
                    priority=getattr(func, "_listener_priority", 100),
                    filters=getattr(func, "_listener_filters", None),
                )
                done = True
            finally:
                if not done:
                    self.unregister_listeners(plug)

    def unregister_listeners(self: "Anjani", plug: plugin.Plugin) -> None:
        for lst in list(self.listeners.values()):
            for listener in lst:
                if listener.plugin == plug:
                    self.unregister_listener(listener)

    async def dispatch_event(
        self: "Anjani",
        event: str,
        *args: Any,
        **kwargs: Any,
    ) -> Optional[Tuple[Any, ...]]:
        results = []

        try:
            listeners = self.listeners[event]
        except KeyError:
            return None

        if not listeners:
            return None

        self.log.debug("Dispatching event '%s' with data %s", event, args)
        EventCount.labels(event).inc()
        with EventLatencySecond.labels(event).time():
            match = None
            index = None
            is_tg_event = False
            for lst in listeners:
                if lst.filters:
                    for idx, arg in enumerate(args):
                        is_tg_event = isinstance(arg, EventType)
                        if is_tg_event:
                            if not await lst.filters(self.client, arg):
                                continue

                            match = arg.matches
                            index = idx
                            break

                        self.log.error(f"'{type(arg)}' can't be used with filters.")
                    else:
                        continue

                if match and index is not None:
                    args[index].matches = match

                result = None
                try:
                    result = await lst.func(*args, **kwargs)
                except KeyError:
                    continue
                except StopPropagation:
                    break
                except Exception as err:  # skipcq: PYL-W0703
                    UnhandledError.labels("command").inc()
                    dispatcher_error = EventDispatchError(
                        f"raised from {type(err).__name__}: {str(err)}"
                    ).with_traceback(err.__traceback__)
                    if is_tg_event and args[0] is not None:
                        data = _get_event_data(args[0])
                        self.log.error(
                            "Error dispatching event '%s' on %s\n"
                            "  Data:\n"
                            "    • Chat    -> %s (%d)\n"
                            "    • Invoker -> %s (%d)\n"
                            "    • Input   -> %s",
                            event,
                            lst.func.__qualname__,
                            data.get("chat_title", "Unknown"),
                            data.get("chat_id", -1),
                            data.get("user_name", "Unknown"),
                            data.get("user_id", -1),
                            data.get("input"),
                            exc_info=dispatcher_error,
                        )
                        await self.dispatch_alert(
                            f"Event __{event}__ on `{lst.func.__qualname__}`",
                            dispatcher_error,
                            data.get("chat_id"),
                        )
                    else:
                        self.log.error(
                            "Error dispatching event '%s' on %s with data\n%s",
                            event,
                            lst.func.__qualname__,
                            _unpack_args(args),
                            exc_info=dispatcher_error,
                        )
                        await self.dispatch_alert(
                            f"Event __{event}__ on `{lst.func.__qualname__}`",
                            dispatcher_error,
                        )
                    continue
                finally:
                    if result:
                        results.append(result)

                    match = None
                    index = None
                    result = None

            return tuple(results)

    async def dispatch_missed_events(self: "Anjani") -> None:
        if not self.loaded or self._TelegramBot__running:
            return

        collection = self.db.get_collection("SESSION")

        data = await collection.find_one(
            {"_id": sha256(self.config.BOT_TOKEN.encode()).hexdigest()}
        )
        if not data:
            return

        pts, date = data.get("pts"), data.get("date")
        if not pts or not date:
            return

        async def send_missed_message(
            messages: list[raw.base.message.Message],
            users: MutableMapping[int, Any],
            chats: MutableMapping[int, Any],
        ) -> None:
            for message in messages:
                self.log.debug("Sending missed message with data '%s'", message)
                await self.client.dispatcher.updates_queue.put(
                    (
                        raw.types.update_new_message.UpdateNewMessage(
                            message=message, pts=0, pts_count=0
                        ),
                        users,
                        chats,
                    )
                )

        async def send_missed_update(
            updates: list[raw.base.update.Update],
            users: MutableMapping[int, Any],
            chats: MutableMapping[int, Any],
        ) -> None:
            for update in updates:
                self.log.debug("Sending missed update with data '%s'", update)
                await self.client.dispatcher.updates_queue.put((update, users, chats))

        try:
            while True:
                # TO-DO
                # 1. Change qts to 0, because we want to get all missed events
                #    so we have a proper loop going on until DifferenceEmpty
                diff = await self.client.invoke(
                    functions.updates.get_difference.GetDifference(pts=pts, date=date, qts=-1)
                )
                if isinstance(
                    diff,
                    (
                        raw.types.updates.difference.Difference,
                        raw.types.updates.difference_slice.DifferenceSlice,
                    ),
                ):
                    if isinstance(diff, raw.types.updates.difference.Difference):
                        state: Any = diff.state
                    else:
                        state: Any = diff.intermediate_state

                    pts, date = state.pts, state.date
                    users = {u.id: u for u in diff.users}  # type: ignore
                    chats = {c.id: c for c in diff.chats}  # type: ignore

                    await asyncio.wait(
                        (
                            self.loop.create_task(
                                send_missed_message(diff.new_messages, users, chats)
                            ),
                            self.loop.create_task(
                                send_missed_update(diff.other_updates, users, chats)
                            ),
                        )
                    )
                elif isinstance(diff, raw.types.updates.difference_empty.DifferenceEmpty):
                    self.log.info("Missed event exhausted, you are up to date.")
                    date = diff.date
                    break
                elif isinstance(diff, raw.types.updates.difference_too_long.DifferenceTooLong):
                    pts = diff.pts
                    continue
                else:
                    break
        except (OSError, asyncio.CancelledError):
            pass
        finally:
            # Unset after we finished to avoid sending the same pts and date,
            # If GetState() doesn't executed on stop event
            await collection.update_one(
                {"_id": sha256(self.config.BOT_TOKEN.encode()).hexdigest()},
                {"$unset": {"pts": "", "date": "", "qts": "", "seq": ""}},
            )

    async def dispatch_alert(
        self: "Anjani",
        invoker: str,
        exc: BaseException,
        chat_id: Optional[int] = None,
    ) -> None:
        """Dispatches an alert to the configured alert log."""
        if not self.config.ALERT_LOG:
            return

        log_chat = self.config.ALERT_LOG.split("#")
        log_thread_id = None
        log_chat_id = int(log_chat[0])
        if len(log_chat) == 2:
            log_thread_id = int(log_chat[1])

        alert = f"""🔴 **Anjani ERROR ALERT**

  - **Alert by:** {invoker}
  - **Chat ID:** {chat_id}
  - **Time (UTC):** {datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")}

**ERROR**
```python
{util.error.format_exception(exc)}
```
        """
        await self.client.send_message(
            log_chat_id,
            alert,
            message_thread_id=log_thread_id,  # type: ignore
        )

    async def log_stat(self: "Anjani", stat: str, *, value: int = 1) -> None:
        await self.dispatch_event("stat_listen", stat, value)



"""Anjani Plugin Base"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import logging
import os.path
from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Optional

from typing_extensions import final

from anjani.util.tg import get_text

if TYPE_CHECKING:
    from .core import Anjani


class Plugin:
    # Class variables
    name: ClassVar[str] = "Unnamed"
    disabled: ClassVar[bool] = False
    helpable: ClassVar[bool] = False

    # Instance variables
    bot: "Anjani"
    log: logging.Logger
    comment: Optional[str]

    def __init__(self, bot: "Anjani") -> None:
        self.bot = bot
        self.log = logging.getLogger(type(self).name.lower().replace(" ", "_"))
        self.comment = None

    # {
    # get_text and text can be declare as ClassVar but we don't do it
    # because we want to keep the documentation of the methods.

    @final
    def get_text(
        self,
        chat_id: Optional[int],
        text_name: str,
        *args: Any,
        noformat: bool = False,
        **kwargs: Any,
    ) -> Coroutine[Any, Any, str]:
        """Parse the string with user language setting.

        Parameters:
            chat_id (`int`, *Optional*):
                Id of the sender(PM's) or chat_id to fetch the user language setting.
                If chat_id is None, the language will always use 'en'.
            text_name (`str`):
                String name to parse. The string is parsed from YAML documents.
            *args (`any`, *Optional*):
                One or more values that should be formatted and inserted in the string.
                The value should be in order based on the language string placeholder.
            noformat (`bool`, *Optional*):
                If True, the text returned will not be formated.
                Default to False.
            **kwargs (`any`, *Optional*):
                One or more keyword values that should be formatted and inserted in the string.
                based on the keyword on the language strings.
        """
        return get_text(self.bot, chat_id, text_name, *args, noformat=noformat, **kwargs)

    # Convenient alias get_text method
    @final
    def text(
        self,
        chat_id: Optional[int],
        text_name: str,
        *args: Any,
        noformat: bool = False,
        **kwargs: Any,
    ) -> Coroutine[Any, Any, str]:
        """Parse the string with user language setting.

        Parameters:
            chat_id (`int`, *Optional*):
                Id of the sender(PM's) or chat_id to fetch the user language setting.
                If chat_id is None, the language will always use 'en'.
            text_name (`str`):
                String name to parse. The string is parsed from YAML documents.
            *args (`any`, *Optional*):
                One or more values that should be formatted and inserted in the string.
                The value should be in order based on the language string placeholder.
            noformat (`bool`, *Optional*):
                If True, the text returned will not be formated.
                Default to False.
            **kwargs (`any`, *Optional*):
                One or more keyword values that should be formatted and inserted in the string.
                based on the keyword on the language strings.
        """
        return get_text(self.bot, chat_id or 0, text_name, *args, noformat=noformat, **kwargs)

    # }

    @classmethod
    def format_desc(cls, comment: Optional[str] = None) -> str:
        _comment = comment + " " if comment else ""
        return f"{_comment}plugin '{cls.name}' ({cls.__name__}) from '{os.path.relpath(inspect.getfile(cls))}'"

    def __repr__(self) -> str:
        return "<" + self.format_desc(self.comment) + ">"



"""Spam Prediction plugin"""

# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import re
from hashlib import md5, sha256
from random import randint
from typing import Any, Callable, ClassVar, List, Literal, MutableMapping, Optional, Tuple

from pyrogram.errors import (
    ChatAdminRequired,
    FloodWait,
    MessageDeleteForbidden,
    PeerIdInvalid,
    QueryIdInvalid,
    UserAdminInvalid,
    UserNotParticipant,
)
from pyrogram.types import (
    CallbackQuery,
    Chat,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
)
from pydantic import BaseModel, validator

from anjani import command, filters, listener, plugin, util
from anjani.core.metrics import SpamPredictionStat
from anjani.util.misc import StopPropagation


class TextLanguage(BaseModel):
    language: str
    probability: float


class PredictionResult(BaseModel):
    is_spam: bool
    spam_score: float
    ham_score: float

    @validator("spam_score", "ham_score")
    def calc_score(cls, value: float) -> float:
        return value * 100

    def get_raw(self, field: Literal["ham", "spam"]) -> float:
        return getattr(self, f"{field}_score") / 100


class SpamDetectionResponse(BaseModel):
    prediction: PredictionResult
    language: TextLanguage
    processed_text: str


class SpamPrediction(plugin.Plugin):
    name: ClassVar[str] = "SpamPredict"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    user_db: util.db.AsyncCollection
    setting_db: util.db.AsyncCollection

    __predict_cost: int = 10
    __log_channel: int = -1001314588569

    async def on_load(self) -> None:
        self._api_key = self.bot.config.SPAM_PREDICTION_API
        self._predict_url = self.bot.config.SPAM_PREDICTION_URL

        if not self._api_key or not self._predict_url:
            self.bot.unload_plugin(self)
            return

        self.db = self.bot.db.get_collection("SPAM_DUMP")
        self.user_db = self.bot.db.get_collection("USERS")
        self.setting_db = self.bot.db.get_collection("SPAM_PREDICT_SETTING")

    async def on_chat_migrate(self, message: Message) -> None:
        await self.db.update_one(
            {"chat_id": message.migrate_from_chat_id},
            {"$set": {"chat_id": message.chat.id}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        setting = await self.setting_db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: setting} if setting else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.setting_db.update_one(
            {"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True
        )

    @staticmethod
    def _build_hash(content: str) -> str:
        return sha256(content.strip().encode()).hexdigest()

    def _build_hex(self, id: Optional[int]) -> str:
        if not id:
            id = self.bot.uid
        # skipcq: PTC-W1003
        return md5((str(id) + self.bot.user.username).encode()).hexdigest()  # skipcq: BAN-B324

    async def _collect_random_sample(self, proba: float, uid: Optional[int]) -> None:
        if not uid or uid == self.bot.uid:
            return
        if randint(1, 2) == 2:  # 50% chance to collect a sample
            await self.user_db.update_one(
                {"_id": uid},
                {
                    "$push": {
                        "pred_sample": {
                            "$each": [proba],
                            "$slice": -10,  # Only keep the last 10 samples
                        }
                    }
                },
            )  # Do not upsert

    async def check_spam(self, text: str) -> SpamDetectionResponse:
        async with self.bot.http.post(
            self._predict_url,
            json={"text": text},
            headers={"x-api-key": self._api_key},
        ) as resp:
            if resp.status != 200:
                raise ValueError(f"Failed to get prediction: {resp.status}")
            res = await resp.json()
            if not res["data"]:
                raise ValueError("Unexpected response")

            return SpamDetectionResponse(**res["data"])

    @listener.filters(
        filters.regex(r"spam_check_(?P<value>t|f)") | filters.regex(r"spam_ban_(?P<user>.*)")
    )
    async def on_callback_query(self, query: CallbackQuery) -> None:
        data = query.matches[0].groupdict()
        handler: MutableMapping[str, Callable] = {
            "value": self._spam_vote_handler,
            "user": self._spam_ban_handler,
        }
        for handle in data.keys():
            try:
                await handler[handle](query, data[handle])
            except QueryIdInvalid:
                pass

    async def _spam_ban_handler(self, query: CallbackQuery, user: str) -> None:
        chat = query.message.chat
        try:
            invoker = await chat.get_member(query.from_user.id)
        except UserNotParticipant:
            return await query.answer(
                await self.get_text(chat.id, "error-no-rights"), show_alert=True
            )

        if not invoker.privileges or not invoker.privileges.can_restrict_members:
            return await query.answer(await self.get_text(chat.id, "spampredict-ban-no-perm"))

        keyboard = query.message.reply_markup
        if not isinstance(keyboard, InlineKeyboardMarkup):
            raise ValueError("Reply markup must be an InlineKeyboardMarkup")

        try:
            target = await self.bot.client.get_users(int(user))
        except PeerIdInvalid:
            await query.answer("Error while fetching user!")
            await query.edit_message_reply_markup(
                reply_markup=InlineKeyboardMarkup(keyboard.inline_keyboard[:-1])
            )
            return
        if isinstance(target, list):
            target = target[0]

        await chat.ban_member(target.id)
        await query.answer(
            await self.get_text(
                chat.id,
                "spampredict-ban",
                user=util.tg.get_username(target) or target.first_name,
            )
        )

        await query.edit_message_reply_markup(
            reply_markup=InlineKeyboardMarkup(keyboard.inline_keyboard[:-1])
        )

    async def _spam_vote_handler(self, query: CallbackQuery, value: str) -> None:
        message = query.message
        content = re.compile(r"([A-Fa-f0-9]{64})").search(message.text)
        author = query.from_user.id

        if not content:
            return await query.answer(f"Can't get hash from MessageID: '{message.id}'")

        content_hash = content[0]

        data = await self.db.find_one({"_id": content_hash})
        if not data:
            return await query.answer(
                "The voting poll for this message has ended!", show_alert=True
            )

        users_on_correct = data["spam"]
        users_on_incorrect = data["ham"]
        if value == "t":
            try:
                # Check user in incorrect data
                if author in users_on_incorrect:
                    users_on_incorrect.remove(author)

                if author in users_on_correct:
                    users_on_correct.remove(author)
                    ans = "You have unvoted this message as a spam!"
                else:
                    users_on_correct.append(author)
                    ans = "You voted this message as a spam!"
            except TypeError:
                return await query.answer(
                    "You can't vote this anymore, because this was marked as a spam by our staff",
                    show_alert=True,
                )
            await query.answer(ans)
        elif value == "f":
            try:
                # Check user in correct data
                if author in users_on_correct:
                    users_on_correct.remove(author)

                if author in users_on_incorrect:
                    users_on_incorrect.remove(author)
                    ans = "You have unvoted this message as non-spam!"
                else:
                    users_on_incorrect.append(author)
                    ans = "You voted this message as non-spam!"
            except TypeError:
                return await query.answer(
                    "You can't vote this anymore, because this was marked as a spam by our staff",
                    show_alert=True,
                )
            await query.answer(ans)
        else:
            return await query.answer("Invalid keyboard method!", show_alert=True)

        await self.db.update_one(
            {"_id": content_hash}, {"$set": {"spam": users_on_correct, "ham": users_on_incorrect}}
        )

    @listener.filters(filters.group & ~filters.outgoing)
    @listener.priority(70)
    async def on_message(self, message: Message) -> None:
        """Checker service for message"""
        if not await self.is_active(message.chat.id):
            return

        chat = message.chat
        text = (
            message.text
            if message.text
            else (message.caption if message.media and message.caption else None)
        )
        if not chat or message.left_chat_member:
            return

        if not text:
            return

        # Always check the spam probability
        await self.spam_check(message, text)

    async def _build_notice(
        self, message: Message, text: str, proba_str: str, identifier: str, content_hash: str
    ) -> Tuple[str, List[List[InlineKeyboardButton]]]:
        notice = (
            "#SPAM_PREDICTION\n\n"
            f"**Prediction Result**: {proba_str}\n"
            f"**Identifier**: `{identifier}`\n"
        )
        if ch := message.forward_from_chat:
            notice += f"**Channel ID**: `{self._build_hex(ch.id)}`\n"

        notice += f"**Message Text Hash**: `{content_hash}`\n\n**====== CONTENT =======**\n\n{text}"

        keyb = [
            [
                InlineKeyboardButton(text="✅ Correct", callback_data="spam_check_t"),
                InlineKeyboardButton(text="❌ Incorrect", callback_data="spam_check_f"),
            ],
            [
                InlineKeyboardButton(
                    text="Chat", url=f"https://t.me/{util.tg.get_username(message.chat)}"
                )
            ],
        ]

        if message.forward_from_chat:
            fw_username = util.tg.get_username(message.forward_from_chat)
            if fw_username:
                raw_btn = InlineKeyboardButton(text="Channel", url=f"https://t.me/{fw_username}")
                keyb.append([raw_btn])

        return notice, keyb

    async def spam_check(self, message: Message, text: str) -> None:
        text = text.strip()
        try:
            user = message.from_user.id
        except AttributeError:
            user = None

        try:
            result = await self.check_spam(text)
        except ValueError:
            self.bot.log.debug("Failed to get prediction")
            return

        await self.bot.log_stat("predicted")
        SpamPredictionStat.labels("predicted").inc()

        probability = result.prediction.spam_score

        await self._collect_random_sample(result.prediction.get_raw("spam"), user)

        if probability <= 50:
            return

        content_hash = self._build_hash(text)
        identifier = self._build_hex(user)
        proba_str = str(probability)
        msg_id = None

        # only log public chat
        if util.tg.get_username(message.chat):
            notice, keyb = await self._build_notice(
                message, text, proba_str, identifier, content_hash
            )

            async with asyncio.Lock():
                data = await self.db.find_one({"_id": content_hash})
                if data:
                    msg_id = data["msg_id"]
                else:
                    while True:
                        try:
                            msg = await self.bot.client.send_message(
                                chat_id=self.__log_channel,
                                text=notice,
                                disable_web_page_preview=True,
                                reply_markup=InlineKeyboardMarkup(keyb),
                            )
                            msg_id = msg.id
                        except FloodWait as flood:
                            await asyncio.sleep(flood.value)  # type: ignore
                            continue

                        await asyncio.sleep(0.1)
                        break

                    await self.db.insert_one(
                        {
                            "_id": content_hash,
                            "user": identifier,
                            "spam": [],
                            "ham": [],
                            "proba": probability,
                            "msg_id": msg.id,
                            "date": util.time.sec(),
                            "text": result.processed_text,
                        },
                    )

        if probability >= 78:
            chat = message.chat
            if not user and message.sender_chat:
                if message.sender_chat.id == chat.id:  # anon admin
                    return

                current_chat: Chat = await self.bot.get_chat(chat.id)
                if (
                    current_chat.linked_chat
                    and message.sender_chat.id == current_chat.linked_chat.id
                ):
                    # Linked channel group
                    return

            target = None
            if user:
                try:
                    target = await message.chat.get_member(user)
                except (UserNotParticipant, ChatAdminRequired):
                    pass
                else:
                    if util.tg.is_staff_or_admin(target):
                        return

            alert = (
                f"❗️**MESSAGE SPAM ALERT**❗️\n\n"
                f"**User**: `{identifier}`\n"
                f"**Message Text Hash**: `{content_hash}`\n"
                f"**Spam Probability**: `{proba_str}%`"
            )

            await self.bot.log_stat("spam_detected")
            SpamPredictionStat.labels("detected").inc()
            try:
                await message.delete()
            except (MessageDeleteForbidden, ChatAdminRequired, UserAdminInvalid):
                alert += "\n\n⚠️Not enough permission to delete message."
                reply_id = message.id
            else:
                await self.bot.log_stat("spam_deleted")
                SpamPredictionStat.labels("deleted").inc()
                alert += "\n\nThe message has been deleted."
                reply_id = 0

            chat = message.chat
            button = []
            if util.tg.get_username(message.chat) and msg_id:
                button.append(
                    [
                        InlineKeyboardButton(
                            "View Message", url=f"https://t.me/SpamPredictionLog/{msg_id}"
                        )
                    ]
                )

            if target is not None:
                me = await chat.get_member(self.bot.uid)
                if me.privileges and me.privileges.can_restrict_members:
                    button.append(
                        [
                            InlineKeyboardButton(
                                "Ban User (*admin only)", callback_data=f"spam_ban_{user}"
                            )
                        ]
                    )

            await self.bot.client.send_message(
                chat.id,
                alert,
                reply_to_message_id=reply_id,
                reply_markup=InlineKeyboardMarkup(button),
                message_thread_id=message.message_thread_id,
            )
            raise StopPropagation

    @command.filters(filters.staff_only)
    async def cmd_spam(self, ctx: command.Context) -> Optional[str]:
        """Manual spam detection by bot staff"""
        user_id = None
        reply_msg = ctx.msg.reply_to_message
        if reply_msg:
            content = reply_msg.text or reply_msg.caption
            if reply_msg.from_user and reply_msg.from_user.id != ctx.author.id:
                user_id = reply_msg.from_user.id
            elif reply_msg.forward_from:
                user_id = reply_msg.forward_from.id
        else:
            content = ctx.input

        if not content:
            return await ctx.get_text("spampredict-empty")

        identifier = self._build_hex(user_id)
        content_hash = self._build_hash(content)
        try:
            prediction = await self.check_spam(content)
        except ValueError:
            return "Prediction failed"

        proba = prediction.prediction.spam_score
        text = f"#SPAM\n\n**CPU Prediction**: `{proba}`\n"
        if identifier:
            text += f"**Identifier**: `{identifier}`\n"

        text += f"**Message Hash**: `{content_hash}`\n\n**======= CONTENT =======**\n\n{content}"
        _, msg, __, ___ = await asyncio.gather(
            self.db.update_one(
                {"_id": content_hash},
                {
                    "$set": {
                        "text": prediction.processed_text,
                        "spam": 1,
                        "ham": 0,
                    }
                },
                upsert=True,
            ),
            self.bot.client.send_message(
                chat_id=self.__log_channel,
                text=text,
                disable_web_page_preview=True,
            ),
            self.bot.log_stat("spam_detected"),
            self.bot.log_stat("predicted"),
        )
        await ctx.respond(
            "Message logged as spam!",
            reply_markup=InlineKeyboardMarkup(
                [[InlineKeyboardButton("View Message", url=msg.link)]]
            ),
        )
        return None

    @command.filters(aliases=["prediction"])
    async def cmd_predict(self, ctx: command.Context) -> Optional[str]:
        """Look a prediction for a replied message"""
        chat = ctx.chat
        user = await self.user_db.find_one({"_id": ctx.author.id})
        if not user:
            return None

        if user.get("reputation", 0) < self.__predict_cost:
            return await self.text(
                chat.id,
                "spampredict-insuficent",
                self.__predict_cost,
                user.get("reputation", 0),
            )

        replied = ctx.msg.reply_to_message
        if not replied:
            await ctx.respond(await ctx.get_text("error-reply-to-message"), delete_after=5)
            return None

        content = replied.text or replied.caption
        if not content:
            return await ctx.get_text("spampredict-empty")
        try:
            prediction = await self.check_spam(content)
        except ValueError:
            return await ctx.get_text("spampredict-failed")

        await self.bot.log_stat("predicted")

        textPrediction = (
            f"**Is Spam**: {prediction.prediction.is_spam}\n"
            f"**Spam Prediction**: `{prediction.prediction.spam_score}`\n"
            f"**Ham Prediction**: `{prediction.prediction.ham_score}`"
        )
        await asyncio.gather(
            self.bot.log_stat("predicted"),
            ctx.respond(
                "**Result**\n\n" + textPrediction,
                reply_to_message_id=replied.id,
            ),
            self.user_db.update_one(
                {"_id": ctx.author.id}, {"$inc": {"reputation": -self.__predict_cost}}
            ),
        )
        return None

    async def setting(self, chat_id: int, setting: bool) -> None:
        """Turn on/off spam prediction in chats"""
        await self.setting_db.update_one(
            {"chat_id": chat_id}, {"$set": {"setting": setting}}, upsert=True
        )

    async def is_active(self, chat_id: int) -> bool:
        """Return SpamShield setting"""
        data = await self.setting_db.find_one({"chat_id": chat_id})
        return data.get("setting", True) if data else True

    @command.filters(filters.admin_only, aliases=["spampredict", "spam_predict"])
    async def cmd_spam_prediction(self, ctx: command.Context, enable: Optional[bool] = None) -> str:
        """Set spam prediction setting"""
        chat = ctx.chat
        if not ctx.input:
            return await self.text(chat.id, "spampredict-view", await self.is_active(chat.id))

        if enable is None:
            return await self.text(chat.id, "err-invalid-option")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "spampredict-set", "on" if enable else "off"),
            self.setting(chat.id, enable),
        )
        return ret



"""Anjani base database"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, List, Literal, Mapping, Optional, Union

from bson.codec_options import CodecOptions
from bson.dbref import DBRef
from bson.son import SON
from bson.timestamp import Timestamp
from pymongo.collation import Collation
from pymongo.database import Database
from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern

from anjani import util

from .base import AsyncBaseProperty
from .change_stream import AsyncChangeStream
from .client_session import AsyncClientSession
from .collection import AsyncCollection
from .command_cursor import AsyncCommandCursor, AsyncLatentCommandCursor, CommandCursor
from .typings import ReadPreferences

if TYPE_CHECKING:
    from .client import AsyncClient


class AsyncDatabase(AsyncBaseProperty):
    """AsyncIO :obj:`~Database`

    *DEPRECATED* methods are removed in this class.
    """

    _client: "AsyncClient"

    dispatch: Database

    def __init__(self, client: "AsyncClient", database: Database) -> None:
        self._client = client

        # Propagate initialization to base
        super().__init__(database)

    def __bool__(self) -> bool:
        return self.dispatch is not None

    def __getitem__(self, name) -> AsyncCollection:
        return AsyncCollection(self, name)

    def __hash__(self) -> int:
        return hash((self.client, self.name))

    def aggregate(
        self,
        pipeline: List[Mapping[str, Any]],
        *args: Any,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> AsyncLatentCommandCursor:
        return AsyncLatentCommandCursor(
            self["$cmd.aggregate"],
            self.dispatch.aggregate,
            pipeline,
            session=session.dispatch if session else session,
            *args,
            **kwargs,
        )

    async def command(
        self,
        command: Union[str, Mapping[str, Any]],
        *,
        value: int = 1,
        check: bool = True,
        allowable_errors: Optional[str] = None,
        read_preference: Optional[ReadPreferences] = None,
        codec_options: Optional[CodecOptions] = None,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.command,
            command,
            value=value,
            check=check,
            allowable_errors=allowable_errors,
            read_preference=read_preference,
            codec_options=codec_options,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def close(self) -> None:
        await self._client.close()

    async def create_collection(
        self,
        name: str,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
        session: Optional[AsyncClientSession] = None,
        check_exists: bool = True,
        **kwargs: Any,
    ) -> AsyncCollection:
        return AsyncCollection(
            self,
            name,
            collection=await util.run_sync(
                self.dispatch.create_collection,
                name,
                codec_options=codec_options,
                read_preference=read_preference,
                write_concern=write_concern,
                read_concern=read_concern,
                session=session.dispatch if session else session,
                check_exists=check_exists,
                **kwargs,
            ),
            session=session,
        )

    async def dereference(
        self, dbref: DBRef, *, session: Optional[AsyncClientSession] = None, **kwargs: Any
    ) -> Optional[Mapping[str, Any]]:
        return await util.run_sync(
            self.dispatch.dereference,
            dbref,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def drop_collection(
        self,
        name_or_collection: Union[str, AsyncCollection],
        session: Optional[AsyncClientSession] = None,
    ) -> Mapping[str, Any]:
        if isinstance(name_or_collection, AsyncCollection):
            name_or_collection = name_or_collection.name

        return await util.run_sync(
            self.dispatch.drop_collection,
            name_or_collection,
            session=session.dispatch if session else session,
        )

    def get_collection(
        self,
        name: str,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
    ) -> AsyncCollection:
        return AsyncCollection(
            self,
            name,
            codec_options=codec_options,
            read_preference=read_preference,
            write_concern=write_concern,
            read_concern=read_concern,
        )

    async def list_collection_names(
        self,
        *,
        session: Optional[AsyncClientSession] = None,
        query: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> List[str]:
        return await util.run_sync(
            self.dispatch.list_collection_names,
            session=session.dispatch if session else session,
            filter=query,
            **kwargs,
        )

    async def list_collections(
        self,
        *,
        session: Optional[AsyncClientSession] = None,
        query: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> AsyncCommandCursor:
        cmd = SON([("listCollections", 1)])
        cmd.update(query, **kwargs)

        res: Mapping[str, Any] = await util.run_sync(
            self.dispatch._retryable_read_command,  # skipcq: PYL-W0212
            cmd,
            session=session.dispatch if session else session,
        )
        return AsyncCommandCursor(CommandCursor(self["$cmd"], res["cursor"], None))

    async def validate_collection(
        self,
        name_or_collection: Union[str, AsyncCollection],
        *,
        scandata: bool = False,
        full: bool = False,
        session: Optional[AsyncClientSession] = None,
        background: Optional[bool] = None,
    ) -> Mapping[str, Any]:
        if isinstance(name_or_collection, AsyncCollection):
            name_or_collection = name_or_collection.name

        return await util.run_sync(
            self.dispatch.validate_collection,
            name_or_collection,
            scandata=scandata,
            full=full,
            session=session.dispatch if session else session,
            background=background,
        )

    def watch(
        self,
        pipeline: Optional[List[Mapping[str, Any]]] = None,
        *,
        full_document: Optional[Literal["updateLookup"]] = None,
        resume_after: Optional[Mapping[str, str]] = None,
        max_await_time_ms: Optional[int] = None,
        batch_size: Optional[int] = None,
        collation: Optional[Collation] = None,
        start_at_operation_time: Optional[Timestamp] = None,
        session: Optional[AsyncClientSession] = None,
        start_after: Optional[Mapping[str, str]] = None,
        comment: Optional[str] = None,
        full_document_before_change: Optional[Literal["required", "whenAvailable"]] = None,
    ) -> AsyncChangeStream:
        return AsyncChangeStream(
            self,
            pipeline,
            full_document,
            resume_after,
            max_await_time_ms,
            batch_size,
            collation,
            start_at_operation_time,
            session,
            start_after,
            comment,
            full_document_before_change,
        )

    def with_options(
        self,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
    ) -> "AsyncDatabase":
        self.dispatch = self.dispatch.with_options(
            codec_options=codec_options,
            read_preference=read_preference,
            write_concern=write_concern,
            read_concern=read_concern,
        )

        return self

    @property
    def client(self) -> "AsyncClient":
        return self._client

    @property
    def name(self) -> str:
        return self.dispatch.name



"""Anjani converter"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from functools import partial
from types import FunctionType
from typing import Any, Dict, List, MutableMapping, Optional, Tuple, Type, Union

from pyrogram import types
from pyrogram.client import Client
from pyrogram.errors import PeerIdInvalid

from anjani.command import CommandFunc, Context
from anjani.error import BadArgument, BadBoolArgument, BadResult, ConversionError

__all__ = [
    "Converter",
    "UserConverter",
    "ChatConverter",
    "ChatMemberConverter",
    "parse_arguments",
]


class Converter:
    """Base class of custom converters that require the `~Context` to be passed.

    Class that derived this base converter need to have the `__call__`
    to do the conversion. This method should also be a `coroutine`.
    """

    async def __call__(self, ctx: Context, arg: str) -> None:  # skipcq: PYL-W0613
        """The base method that should be overrided and will be called on conversion.

        Parameters:
            ctx (`~Context`): The command invocation context that the argument used in.
            arg (`str`): The argument that is being converted.

        Raises:
            ConversionError: The converter failed to convert an argument.
        """
        raise NotImplementedError("Derived classes need to implement __call__ method!")


class EntityConverter(Converter):  # skipcq: PYL-W0223
    @staticmethod
    def parse_entities(msg: types.Message, arg: str) -> Optional[types.User]:
        for i in msg.entities:
            if i.type == "text_mention" and msg.text[i.offset : i.offset + i.length] == arg:
                return i.user
        return None


class UserConverter(EntityConverter):
    """Converts to a `~pyrogram.types.User`.

    Conversion priority:
    1. By user id.
    2. By username.
    3. By text mention.
    """

    async def extract_user(self, client: Client, user_id: Union[str, int]) -> types.User:
        """Excract user from user id"""
        try:
            user = await client.get_users(user_id)
            if isinstance(user, types.User):
                return user

            raise BadResult(f"Invalid conversion types '{type(user)}' result")
        except PeerIdInvalid as err:
            raise ConversionError(self, err) from err

    async def __call__(self, ctx: Context, arg: str) -> Optional[types.User]:
        if arg.isdigit() or arg.startswith("@"):
            try:
                user = await ctx.bot.client.get_users(arg)
                if isinstance(user, types.User):
                    return user

                raise BadResult(f"Invalid conversion types '{type(user)}' result")
            except PeerIdInvalid as err:
                raise ConversionError(self, err) from err
        return self.parse_entities(ctx.msg, arg)


class ChatConverter(Converter):
    """Converts to a `~pyrogram.types.ChatMember`.

    Conversion priority:
    1. By chat id.
    2. By chat username (with '@').
    3. By chat username (without '@').
    """

    async def __call__(self, ctx: Context, args: str) -> types.Chat:
        try:
            chat = await ctx.bot.client.get_chat(args)
            if isinstance(chat, types.Chat):
                return chat

            raise BadResult(f"Invalid conversion types `{type(chat)}` result")
        except PeerIdInvalid as err:
            raise ConversionError(self, err) from err


class ChatMemberConverter(EntityConverter):
    """Converts to a `~pyrogram.types.ChatMember`.

    Conversion priority:
    1. By user id.
    2. By username.
    3. By text mention.
    """

    async def get_member(
        self, client: Client, chat_id: int, user_id: Union[int, str]
    ) -> types.ChatMember:
        try:
            return await client.get_chat_member(chat_id, user_id)
        except PeerIdInvalid as err:
            raise ConversionError(self, err) from err

    async def __call__(self, ctx: Context, arg: str) -> Optional[types.ChatMember]:
        if arg.isdigit() or arg.startswith("@"):
            return await self.get_member(ctx.bot.client, ctx.chat.id, arg)
        res = self.parse_entities(ctx.msg, arg)
        if res:
            return await self.get_member(ctx.bot.client, ctx.chat.id, res.id)

        return None


CONVERTER_MAP: MutableMapping[Type[Any], Any] = {
    types.User: UserConverter,
    types.Chat: ChatConverter,
    types.ChatMember: ChatMemberConverter,
}


def _bool_converter(arg: str) -> Union[bool, BadBoolArgument]:
    arg = arg.lower()
    if arg in ("yes", "true", "enable", "on", "1"):
        return True
    if arg in ("no", "false", "disable", "off", "0"):
        return False
    raise BadBoolArgument(f"Unrecognized argument of boolean '{arg}'")


def _get_default(param: inspect.Parameter, default: Any = None) -> Union[Any, None]:
    return param.default if param.default is not param.empty else default


async def transform(ctx: Context, param: inspect.Parameter, arg: str) -> Any:
    converter = param.annotation

    if converter is param.empty:
        return arg

    # Check if the annotation was an `Optional` or `Union` type.
    # This type hinting make a parsing ambiguities.
    # Hence we just simply use the first arg as the converter if is not None type.
    # Else use the second arg.
    if getattr(converter, "__origin__", None) is Union:
        if converter.__args__[0] is None:
            converter = converter.__args__[1]
        else:
            converter = converter.__args__[0]

    if isinstance(converter, (FunctionType, partial)):
        if inspect.iscoroutinefunction(converter):
            return await converter(arg)

        return converter(arg)

    try:
        module = converter.__module__
    except AttributeError:
        pass
    else:
        if module is not None and module.startswith("pyrogram."):
            converter = CONVERTER_MAP.get(converter, converter)

    if inspect.isclass(converter) and issubclass(converter, Converter):
        try:
            return await converter()(ctx, arg)
        except ConversionError as err:
            return _get_default(param, err)

    if converter is bool:
        try:
            return _bool_converter(arg)
        except BadBoolArgument as err:
            return _get_default(param, err)

    try:
        return converter(arg)
    except ValueError as err:
        return _get_default(param, err)


async def parse_arguments(
    sig: inspect.Signature, ctx: Context, func: CommandFunc
) -> Tuple[List[Any], Dict[Any, Any]]:
    to_convert = ctx.args
    args = []  # type: List[Any]
    kwargs = {}  # type: Dict[Any, Any]
    idx = 0
    items = iter(sig.parameters.items())

    # skip Context argument
    next(items)  # skipcq: PTC-W0063
    for name, param in items:
        if param.kind in (param.POSITIONAL_OR_KEYWORD, param.POSITIONAL_ONLY):
            try:
                result = await transform(ctx, param, to_convert[idx])
            except IndexError:
                result = _get_default(param)
            args.append(result)
            idx += 1
        elif param.kind == param.KEYWORD_ONLY:
            # Consume remaining text to the kwargs
            kwargs[name] = " ".join(to_convert[idx:]).strip()
            break
        elif param.kind in {param.VAR_POSITIONAL, param.VAR_KEYWORD}:
            raise BadArgument(
                f"Unsuported {param.kind} parameter conversion "
                f"Found '*{name}' on '{func.__name__}'"
            )
    return args, kwargs
